function [projectedI] = WeakRandPOQCS(tvI,dctQCoefs,Q)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% -------------------------------------------------------------------------
% Copyright (c) 2013 GIPSA-Lab/Grenoble INP, and Beihang University
% All Rights Reserved.
% -------------------------------------------------------------------------
% Permission to use, copy, or modify this software and its documentation
% for educational and research purposes only and without fee is hereby
% granted, provided that this copyright notice and the original authors'
% names appear on all copies and supporting documentation. This program
% shall not be used, rewritten, or adapted as the basis of a commercial
% software or hardware product without first obtaining permission of the
% authors. The authors make no representations about the suitability of
% this software for any purpose. It is provided "as is" without express
% or implied warranty.
% -------------------------------------------------------------------------
% If you find any bugs, please kindly report to us.
% -------------------------------------------------------------------------
% 
% description:   randomly project DCT coefficients to their original 
%                quantization bin when they are outside the original or the
%                neighobing quantization bins during the TV-based 
%                deblocking procedure
% 
% requires:      Matlab JPEG Toolbox available at:
%                http://www.philsallee.com/jpegtbx/index.html
% 
% INPUT
%           tvI: image pixel value matrix 
%     dctQCoefs: DCT coefficients of the JPEG image
%             Q: quantization table, 8 * 8 sized
% 
% OUTPUT
%    projectedI: image matrix after the weak random projection
% 
% paper:         W. Fan, K. Wang, F. Cayre, and Z. Xiong,
%                "A variational approach to JPEG anti-forensics",
%                ICASSP, 2013.
% 
% contact:       wei.fan@gipsa-lab.grenoble-inp.fr
% last modified: Dec. 6th, 2013
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

n = 8; % DCT block size
[nH,nW] = size(tvI); % image size

projectedICoefs = bdct(double(tvI)-128); % transform to DCT-domain
Qmat = repmat(Q,nH/n,nW/n); % quantization table

% outliers
outliers = round(projectedICoefs./Qmat) < quantize(dctQCoefs,Q)-1 | ...
           round(projectedICoefs./Qmat) > quantize(dctQCoefs,Q)+1;
       
% rand noise to be added to DCT coefficients
randN_pos = (2*abs(rand(nH,nW)-0.5) - 0.5).*Qmat; % half closed interval [-Q/2,Q/2)
randN_zero = rand(nH,nW).*Qmat./2; % open interval (-Q/2,Q/2)
rand_neg = (-2*abs(rand(nH,nW)-0.5) + 0.5).*Qmat; % half closed interval (-Q/2,Q/2]

% randomly project the outliers to the orginal quantization bins
projectedICoefs(outliers&dctQCoefs>0) = dctQCoefs(outliers&dctQCoefs>0) + randN_pos(outliers&dctQCoefs>0);
projectedICoefs(outliers&dctQCoefs==0) = dctQCoefs(outliers&dctQCoefs==0) + randN_zero(outliers&dctQCoefs==0);
projectedICoefs(outliers&dctQCoefs<0) = dctQCoefs(outliers&dctQCoefs<0) + rand_neg(outliers&dctQCoefs<0);

% transform back to spatial-domain
projectedI = ibdct(projectedICoefs)+128;
projectedI = double(uint8(projectedI)); % rounding & truncation

end